from typing import Any, Dict, List

from sktmls.models import MLSModelError, MLSRuleModel


class VasXcloudRuleModel(MLSRuleModel):
    """
    ## Args

    - model_name: (str) 모델 이름
    - model_version: (str) 모델 버전
    - features: (list(str)) 피쳐 리스트

    ## Example

    ```
    my_model = VasXcloudRuleModel(
        model_name="vas_xloud_rule_model",
        model_version="v1",
        features=[
            "age",
            "real_arpu_bf_m1",
            "app_use_traffic_game",
            "app_use_days_video_median_yn",
            "data_use_night_ratio_median_yn",
        ],
    )

    result = my_model.predict([20, 55000, 10000, "Y", "Y"])

    ```
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if len(self.features) != len(x):
            raise MLSModelError("The length of input is different from that of features in model.json")

        age = x[0]
        real_arpu_bf_m1 = x[1]
        app_use_traffic_game = x[2]
        app_use_days_video_median_yn = x[3]
        data_use_night_ratio_median_yn = x[4]

        if (
            (age >= 19 and age <= 55)
            and real_arpu_bf_m1 >= 30000
            and (
                app_use_traffic_game > 0 or app_use_days_video_median_yn == "Y" or data_use_night_ratio_median_yn == "Y"
            )
        ):
            return {
                "items": [
                    {
                        "id": "NA00007114",
                        "name": "게임패스 얼티밋",
                        "type": "vas_xcloud",
                        "props": {"context_id": "context_default"},
                    }
                ]
            }
        else:
            return {"items": []}



from typing import List
from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSClientError

MLS_DATA_STATUS_API_URL = "/api/v1/data_status"


class S3Status:
    """
    MLS S3 업로드 상태를 나타내는 클래스입니다.
    """

    def __init__(self, table: str, daily_status: List[dict]):
        """
        ## Args

        - kwargs
            - table: (str) 테이블 명
            - daily_status: (list) 최근 5일간의 상태 값 리스트

        ## Returns
        `sktmls.data_status.S3Status`
        """
        self.table = table
        self.daily_status = daily_status

    def get(self):
        return self.__dict__


class BatchStatus:
    """
    MLS Batch 작업 상태를 나타내는 클래스입니다.
    """

    def __init__(self, job_status: str, jobs: List[dict]):
        """
        ## Args

        - kwargs
            - job_status: (str) 작업 상태 (`running`|`success`|`fail`)
            - jobs: (list) 작업 리스트

        ## Returns
        `sktmls.data_status.BatchStatus`
        """
        self.job_status = job_status
        self.jobs = jobs

    def get(self):
        return self.__dict__


class DynamoDBStatus:
    """
    MLS DynamoDB 업로드 상태를 나타내는 클래스입니다.
    """

    def __init__(self, table: str, status: List[dict]):
        """
        ## Args

        - kwargs
            - table: (str) 테이블 명
            - status: (list) 최근 5일간의 상태값 리스트

        ## Returns
        `sktmls.data_status.DynamoDBStatus`
        """
        self.table = table
        self.status = status

    def get(self):
        return self.__dict__


class StatusClient(MLSClient):
    """
    MLS Status 관련 기능들을 제공하는 클라이언트입니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
        username: str = None,
        password: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.data_status.StatusClient`

        ## Example

        ```
        status_client = StatusClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def list_s3_status(self) -> List[S3Status]:
        """
        S3 상태 리스트를 가져옵니다.

        ## Returns
        list(`sktmls.data_status.S3Status`)

        - table: (str) 테이블 명
        - daily_status: (list(dict)) 최근 5일간의 상태 값 리스트

        ## Example

        ```
        s3_status_list = status_client.list_s3_status()
        ```
        """
        return [
            S3Status(table=table, daily_status=daily_status)
            for table, daily_status in self._request(method="GET", url=f"{MLS_DATA_STATUS_API_URL}/s3")
            .results["s3"]
            .items()
        ]

    def get_s3_status(self, table: str) -> S3Status:
        """
        S3 상태 정보를 가져옵니다.

        ## Args

        - table: (str) 테이블 명

        ## Returns
        `sktmls.data_status.S3Status`

        - table: (str) 테이블 명
        - daily_status: (list(dict)) 최근 5일간의 상태 값 리스트

        ## Example

        ```
        s3_status = status_client.get_s3_status(table="sample_table")
        ```
        """
        try:
            daily_status = self._request(method="GET", url=f"{MLS_DATA_STATUS_API_URL}/s3").results["s3"][table]
        except KeyError:
            raise MLSClientError(code=404, msg="해당 테이블에 대한 상태값이 없습니다")

        return S3Status(
            table=table,
            daily_status=daily_status,
        )

    def list_batch_status(self) -> List[BatchStatus]:
        """
        Batch 상태 리스트를 가져옵니다.

        ## Returns
        list(`sktmls.data_status.BatchStatus`)

        - job_status: (str) 작업 상태 (`running`|`success`|`fail`)
        - jobs: (list(dict)) 작업 리스트

        ## Example

        ```
        batch_status_list = status_client.list_batch_status()
        ```
        """
        return [
            BatchStatus(job_status=job_status, jobs=jobs)
            for job_status, jobs in self._request(method="GET", url=f"{MLS_DATA_STATUS_API_URL}/batch")
            .results["batch"]
            .items()
        ]

    def get_batch_status(self, job_status: str) -> BatchStatus:
        """
        Batch 상태를 가져옵니다.

        ## Args

        - job_status: (str) 작업 상태 (`running`|`success`|`fail`)

        ## Returns
        `sktmls.data_status.BatchStatus`

        - job_status: (str) 작업 상태 (`running`|`success`|`fail`)
        - jobs: (list(dict)) 작업 리스트

        ## Example

        ```
        batch_status = status_client.get_batch_status(job_status="success")
        ```
        """
        assert job_status in ["running", "success", "fail"], "작업의 상태는 running, success, fail 중 하나여야합니다"

        return BatchStatus(
            job_status=job_status,
            jobs=self._request(method="GET", url=f"{MLS_DATA_STATUS_API_URL}/batch").results["batch"][job_status],
        )

    def list_dynamodb_status(self) -> List[DynamoDBStatus]:
        """
        DynamoDB 상태 리스트를 가져옵니다.

        ## Returns
        list(`sktmls.data_status.DynamoDBStatus`)

        - table: (str) 테이블 명
        - status: (list(dict)) 최근 5일간의 상태값 리스트

        ## Example

        ```
        dynamodb_status_list = status_client.list_dynamodb_status()
        ```
        """
        return [
            DynamoDBStatus(table_name, status)
            for table_name, status in self._request(method="GET", url=f"{MLS_DATA_STATUS_API_URL}/dynamodb")
            .results["dynamodb"]
            .items()
        ]

    def get_dynamodb_status(self, table: str) -> DynamoDBStatus:
        """
        DynamoDB 상태를 가져옵니다.

        ## Args

        - table: (str) 테이블 명

        ## Returns
        `sktmls.data_status.DynamoDBStatus`

        - table: (str) 테이블 명
        - status: (list(dict)) 최근 5일간의 상태값 리스트

        ## Example

        ```
        dynamodb_status = status_client.get_dynamodb_status(table="sample_table")
        ```
        """
        try:
            status = self._request(method="GET", url=f"{MLS_DATA_STATUS_API_URL}/dynamodb").results["dynamodb"][table]
        except KeyError:
            raise MLSClientError(code=404, msg="해당 테이블에 대한 상태값이 없습니다")

        return DynamoDBStatus(table=table, status=status)



from .mls_env import MLSENV, MLSRuntimeENV

from .model_registry import ModelRegistry, ModelRegistryError
from .mls_client import MLSClient, MLSResponse, MLSClientError

__all__ = [
    "MLSENV",
    "MLSRuntimeENV",
    "ModelRegistry",
    "ModelRegistryError",
    "MLSClient",
    "MLSResponse",
    "MLSClientError",
]



import random
from datetime import datetime
from typing import Any, Dict, List

from pytz import timezone

from sktmls.models import MLSRuleModel, MLSModelError

TZ = timezone("Asia/Seoul")
SEED_FORMAT = {"daily": "%Y%m%d", "weekly": "%Y%W", "monthly": "%Y%m", "yearly": "%Y"}


class RandomPickModel(MLSRuleModel):
    """
    MLS 모델 레지스트리에 등록되는 Rule 기반 클래스입니다.

    주어진 `candidate` 중 `num_pick`개의 결과를 랜덤하게 추출하여 반환하는 모델입니다.
    """

    def __init__(self, model_name: str, model_version: str, candidates: List[Dict[str, Any]], num_pick: int = 1):
        """
        ## Args

        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - candidates: (list(dict)) 랜덤 추출할 후보 리스트
        - num_pick: (optional) (int) 랜덤 추출할 갯수 (기본값: 1)

        ## Example

        ```
        my_model_v1 = RandomPickModel(
            model_name="my_model",
            model_version="v1",
            candidates=[
                {
                    "id": "item1",
                    "name": "아이템1",
                    "type": "타입",
                    "props": {}
                },
                {
                    "id": "item2",
                    "name": "아이템2",
                    "type": "타입",
                    "props": {}
                },
                {
                    "id": "item3",
                    "name": "아이템3",
                    "type": "타입",
                    "props": {}
                }
            ],
            num_pick=2)

        result = my_model_v1.predict(None)
        ```
        """
        super().__init__(model_name, model_version, ["user_id"])

        if len(candidates) < num_pick:
            raise MLSModelError("The length of num_pick cannot be greater than that of candidates")

        self.candidates = candidates
        self.num_pick = num_pick

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        return {"items": random.sample(self.candidates, self.num_pick)}


class PeriodicRandomPickModel(RandomPickModel):
    """
    MLS 모델 레지스트리에 등록되는 Rule 기반 클래스입니다.

    주어진 `candidate` 중 `num_pick`개의 결과를 랜덤하게 추출하여 반환하는 모델입니다.

    `random_cycle` 값을 참조하여 랜덤 시드를 설정합니다.
    """

    def __init__(
        self,
        model_name: str,
        model_version: str,
        candidates: List[Dict[str, Any]],
        num_pick: int = 1,
        random_cycle: str = "daily",
    ):
        """
        ## Args

        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - candidates: (dict) 랜덤 추출할 후보 리스트
        - num_pick: (optional) (int) 랜덤 추출할 갯수 (기본값: 1)
        - random_cycle: (optional) (str) 랜덤 시드 초기화 주기 (`daily`|`weekly`|`monthly`|`yearly`) (기본값: `daily`)

        ## Example

        ```
        my_model_v1 = PeriodicRandomPickModel(
            model_name="my_model",
            model_version="v1",
            candidates=[
                {
                    "id": "item1",
                    "name": "아이템1",
                    "type": "타입",
                    "props": {}
                },
                {
                    "id": "item2",
                    "name": "아이템2",
                    "type": "타입",
                    "props": {}
                },
                {
                    "id": "item3",
                    "name": "아이템3",
                    "type": "타입",
                    "props": {}
                }
            ],
            num_pick=2,
            random_cycle="daily")

        result = my_model_v1.predict("my_svc_mgmt_num")
        ```
        """
        super().__init__(model_name, model_version, candidates, num_pick)

        if random_cycle not in SEED_FORMAT:
            raise MLSModelError("허용되지 않는 `random_cycle`값입니다.")
        self.random_cycle = random_cycle

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        user_id = x[0]
        random.seed(f"{user_id}{datetime.now(TZ).strftime(SEED_FORMAT[self.random_cycle])}")
        return super().predict(x)



from typing import List, Union
from enum import Enum

from dateutil import parser

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSClientError, MLSResponse
from sktmls.datasets import Dataset
from sktmls.components import Component
from sktmls.data_status import StatusClient

MLS_MODELS_API_URL = "/api/v1/models"
MLS_DYNAMODB_TABLES_API_URL = "/api/v1/dynamodb/tables"
MLS_MODEL_REGISTRY_ARTIFACTS_API_URL = "/api/v1/model_registry/artifacts/info"

MODEL_LIBS = ["sklearn", "xgboost", "lightgbm", "rule", "etc"]


class MLModel:
    """
    ML모델 클래스 입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - version: (str) ML모델 버전
            - creator: (str) ML모델 생성 계정명
            - description: (str) ML모델 설명
            - model_type: (str) ML모델 타입 (`automl` | `manual`)
            - table: (str) ML모델 테이블
            - model_data: (str) ML모델 데이터
            - status: (`sktmls.models.MLModelStatus`) ML모델 상태
            - created_at: (datetime) 생성일시
            - updated_at: (datetime) 수정일시

        ## Returns
        `sktmls.models.MLModel`
        """
        self.id = kwargs.get("id")
        self.name = kwargs.get("name")
        self.version = kwargs.get("version")
        self.description = kwargs.get("description")
        self.creator = kwargs.get("creator")
        self.model_type = kwargs.get("model_type")
        self.table = kwargs.get("table")
        self.model_data = kwargs.get("model_data")
        self.status = kwargs.get("status")
        try:
            self.created_at = parser.parse(kwargs.get("created_at"))
        except TypeError:
            self.created_at = None
        try:
            self.updated_at = parser.parse(kwargs.get("updated_at"))
        except TypeError:
            self.updated_at = None

    def __str__(self) -> str:
        return self.name

    def get(self) -> dict:
        return self.__dict__

    def reset(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class AutoMLModel(MLModel):
    """
    AutoML모델 클래스 입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - model_type: (str) ML모델 타입 (`automl` | `manual`)
            - version: (str) ML모델 버전
            - description: (str) ML모델 설명
            - creator: (str) ML모델 생성 계정명
            - status: (`sktmls.models.MLModelStatus`) ML모델 상태
            - model_meta: (dict) ML모델 메타정보
            - dataset_id: (int) ML 데이터셋 고유 ID
            - dataset_name: (str) ML 데이터셋 이름
            - automl_model_info: (dict) AutoML모델 정보

        ## Returns
        `sktmls.models.AutoMLModel`
        """

        super().__init__(**kwargs)
        self.dataset_id = kwargs.get("dataset_id")
        self.dataset_name = kwargs.get("dataset_name")
        self.automl_model_info = kwargs.get("automl_model_info")


class ManualModel(MLModel):
    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - model_type: (str) ML모델 타입 (`automl` | `manual`)
            - version: (str) ML모델 버전
            - description: (str) ML모델 설명
            - creator: (str) ML모델 생성 계정명
            - status: (`sktmls.models.MLModelStatus`) ML모델 상태
            - model_meta: (dict) ML모델 메타정보
            - components: list(`sktmls.components.Component`) ML모델 컴포넌트 리스트
            - table: (str) ML모델 테이블
            - model_lib: (str) ML모델 라이브러리
            - model_data: (str) ML모델 데이터
            - features: (str) ML모델에 사용된 피쳐
            - is_enabled: (bool) 활성상태 여부

        ## Returns
        `sktmls.models.ManualModel`
        """

        super().__init__(**kwargs)
        self.model_meta = kwargs.get("model_meta")
        self.components = (
            [Component(**component) for component in kwargs.get("components")]
            if kwargs.get("components")
            else kwargs.get("components")
        )
        self.model_lib = kwargs.get("model_lib")
        self.features = kwargs.get("features")
        self.is_enabled = kwargs.get("is_enabled")


class MLModelStatus(Enum):
    IN_USE = "IN_USE"
    NOT_IN_USE = "NOT_IN_USE"
    AUTOML_TRAINING = "TRAINING"
    AUTOML_DONE = "DONE"
    AUTOML_FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"


class MLModelClient(MLSClient):
    def __init__(
        self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, username: str = None, password: str = None
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.models.MLModelClient`

        ## Example

        ```
        ml_model_client = MLModelClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def create_automl_model(self, dataset: Dataset, name: str, version: str, description: str) -> AutoMLModel:
        """
        AutoML 모델을 생성합니다.

        ## Args

        - dataset: (sktmls.datasets.Dataset) ML 데이터셋
        - name: (str) 생성할 AutoML 모델 이름
        - version: (str) 생성할 AutoML 모델 버전
        - description: (str) 생성할 AutoML 설명

        ## Returns
        `sktmls.models.AutoMLModel`

        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - creator: (str) ML모델 생성 계정명
        - description: (str) ML모델 설명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - dataset_id: (int) ML 데이터셋 고유 ID
        - dataset_name: (str) ML 데이터셋 이름
        - automl_model_info: (dict) AutoML모델 정보

        ## Example

        ```
        automl_model = ml_model_client.create_automl_model(
            dataset=dataset,
            name="automl_test_model",
            version="v1",
            description='test_model'
        )
        ```
        """
        assert type(dataset) == Dataset

        data = {
            "model_type": "automl",
            "name": name,
            "version": version,
            "description": description,
            "dataset_id": dataset.id,
        }
        response = self._request(method="POST", url=MLS_MODELS_API_URL, data=data).results
        return self.get_automl_model(id=response.get("id"))

    def create_manual_model(
        self,
        name: str,
        version: str,
        description: str = None,
        creator: str = None,
        model_lib: str = None,
        model_data: str = None,
        model_meta: dict = None,
        features: str = None,
        table: str = None,
        components: List[Component] = None,
    ) -> ManualModel:
        """
        일반 모델을 생성합니다. 모델을 버킷에 연결 가능하도록 deploy하는 작업입니다.

        일반 모델은 deploy 후 버킷에 연결되어 정상 동작하기 까지 약 한 시간이 소요됩니다.

        ## Args

        - name: (str) 생성할 일반 모델 이름
        - version: (str) 생성할 일반 모델 버전
        - description: (optional) (str) 생성할 일반 모델 설명
        - creator: (optional) (str) 일반 모델 생성 계정명
        - model_lib: (optional) (str) 일반 모델 라이브러리
        - model_data: (optional) (str) 일반 모델 데이터
        - model_meta: (optional) (dict) 모델 레지스트리에 등록된 메타정보(model.json)
        - features: (optional) (str) 일반 모델에 사용된 피쳐
        - table: (optional) (str) 일반 모델 테이블
        - components: (optional) list(`sktmls.components.Component`) 일반 모델 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```
        component_1 = component_client.get_component(name="test_component_1")
        component_2 = component_client.get_component(name="test_component_2")
        manual_model = ml_model_client.create_manual_model(
            name="manual_test_model",
            version="v1",
            description="test_model",
            creator="mls_user",
            model_lib="lightgbm",
            model_data="/models/manual_test_model/v1/model.joblib",
            model_meta={
                "name": "manual_test_model",
                "version": "v1",
                "model_lib": "lightgbm",
                "model_lib_version": "2.3.1",
                "model_data": "/models/manual_test_model/v1/model.joblib",
                "features": ["age", "app_use_traffic_youtube", "svc_scrb_period"],
            },
            features="age,app_use_traffic_youtube,svc_scrb_period",
            table="test_model_table",
            components=[component_1, component_2]
        )
        ```
        """

        data = {
            "model_type": "manual",
            "name": name,
            "version": version,
        }
        if description:
            data["description"] = description
        if creator:
            data["creator"] = creator
        if model_lib:
            assert model_lib in MODEL_LIBS, "허용되지 않는 `model_lib`값입니다."
            data["model_lib"] = model_lib
        if model_data:
            data["model_data"] = model_data
        if model_meta:
            data["model_meta"] = model_meta
        if features:
            data["features"] = features
        if table:
            data["table"] = table
        if components:
            for component in components:
                assert type(component) == Component, "`components`는 list(`sktmls.components.Component`) 형태여야 합니다."
            data["components"] = [component.name for component in components]

        response = self._request(method="POST", url=MLS_MODELS_API_URL, data=data).results
        return self.get_manual_model(id=response.get("id"))

    def list_automl_models(self, **kwargs) -> List[AutoMLModel]:
        """
        AutoML모델 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 필터 조건
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - version: (str) ML모델 이름

        ## Returns
        list(`sktmls.models.AutoMLModel`)

        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - creator: (str) ML모델 생성 계정명
        - description: (str) ML모델 설명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - dataset_id: (int) ML 데이터셋 고유 ID
        - dataset_name: (str) ML 데이터셋 이름
        - automl_model_info: (dict) AutoML모델 정보

        ## Example

        ```
        automl_models = ml_model_client.list_automl_models()
        ```
        """

        kwargs.update({"model_type": "automl"})
        response = self._request(method="GET", url=MLS_MODELS_API_URL, params=kwargs)
        return [AutoMLModel(**ml_model) for ml_model in response.results]

    def list_manual_models(self, **kwargs) -> List[ManualModel]:
        """
        일반 모델 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 필터 조건
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - version: (str) ML모델 이름

        ## Returns
        list(`sktmls.models.ManualModel`)

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```
        manual_models = ml_model_client.list_manual_models()
        ```
        """
        kwargs.update({"model_type": "manual"})
        response = self._request(method="GET", url=MLS_MODELS_API_URL, params=kwargs)
        return [ManualModel(**ml_model) for ml_model in response.results]

    def get_automl_model(self, id: int = None, name: str = None, version: str = None) -> AutoMLModel:
        """
        AutoML모델 정보를 가져옵니다.

        ## Args: `id` 또는 (`name` and `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - version: (str) ML모델 이름

        ## Returns
        `sktmls.models.AutoMLModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - dataset_id: (int) ML 데이터셋 고유 ID
        - dataset_name: (str) ML 데이터셋 이름
        - automl_model_info: (dict) AutoML모델 정보

        ## Example

        ```
        automl_model_by_id = ml_model_client.get_automl_model(id=3)
        automl_model_by_name_and_version = ml_model_client.get_automl_model(name="my_automl_model", version="my_automl_version")
        ```
        """
        assert id or (name and version), "`id` 또는 (`name`, `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        automl_models = self.list_automl_models(id=id, name=name, version=version)
        if len(automl_models) == 0:
            raise MLSClientError(code=404, msg="ML모델이 없습니다.")

        return AutoMLModel(**self._request(method="GET", url=f"{MLS_MODELS_API_URL}/{automl_models[0].id}").results)

    def get_manual_model(self, id: int = None, name: str = None, version: str = None) -> ManualModel:
        """
        일반 모델 정보를 가져옵니다.

        ## Args: `id` 또는 (`name` and `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - version: (str) ML모델 이름

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```
        manual_model_by_id = ml_model_client.get_manual_model(id=3)
        manual_model_by_name_and_version = ml_model_client.get_manual_model(name="my_manual_model", version="my_manual_version")
        ```
        """
        assert id or (name and version), "`id` 또는 (`name`, `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        manual_models = self.list_manual_models(id=id, name=name, version=version)
        if len(manual_models) == 0:
            raise MLSClientError(code=404, msg="ML모델이 없습니다.")

        return ManualModel(**self._request(method="GET", url=f"{MLS_MODELS_API_URL}/{manual_models[0].id}").results)

    def update_manual_model(
        self,
        manual_model: ManualModel,
        name: str = None,
        version: str = None,
        description: str = None,
        creator: str = None,
        model_lib: str = None,
        model_data: str = None,
        model_meta: dict = None,
        features: str = None,
        table: str = None,
        components: List[Component] = None,
    ) -> ManualModel:
        """
        일반 모델 정보를 수정합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체
        - name: (str) (optional) 수정할 일반 모델 이름
        - version: (str) (optional) 수정할 일반 모델 버전
        - description: (optional) (str) 수정할 일반 모델 설명
        - creator: (optional) (str) 수정할 일반 모델 계정명
        - model_lib: (optional) (str) 수정할 일반 모델 라이브러리
        - model_data: (optional) (str) 수정할 일반 모델 데이터
        - model_meta: (optional) (dict) 수정할 모델 레지스트리에 등록된 메타정보(model.json)
        - features: (optional) (str) 수정할 일반 모델에 사용된 피쳐
        - table: (optional) (str) 수정할 일반 모델 테이블
        - components: (optional) list(`sktmls.components.Component`) 수정할 일반 모델 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```
        manual_model = ml_model_client.get_manual_model(name="my_manual_model", version="v1")
        manual_model = ml_model_client.update_manual_model(
            manual_model=manual_model,
            name="my_new_manual_model",
            description="updated manual model",
            model_data="/models/my_new_manual_model/v1/model.joblib",
        )
        ```
        """
        assert type(manual_model) == ManualModel

        data = {
            "name": manual_model.name,
            "version": manual_model.version,
            "model_type": "manual",
        }
        if manual_model.description is not None:
            data["description"] = manual_model.description
        if manual_model.creator is not None:
            data["creator"] = manual_model.creator
        if manual_model.model_lib is not None:
            data["model_lib"] = manual_model.model_lib
        if manual_model.model_data is not None:
            data["model_data"] = manual_model.model_data
        if manual_model.model_meta is not None:
            data["model_meta"] = manual_model.model_meta
        if manual_model.features is not None:
            data["features"] = manual_model.features
        if manual_model.table is not None:
            data["table"] = manual_model.table
        if manual_model.components is not None:
            data["components"] = [component.get("name") for component in manual_model.components]

        if name:
            data["name"] = name
        if version:
            data["version"] = version
        if description is not None:
            data["description"] = description
        if creator is not None:
            data["creator"] = creator
        if model_lib is not None:
            assert model_lib in MODEL_LIBS, "허용되지 않는 `model_lib`값입니다."
            data["model_lib"] = model_lib
        if model_data is not None:
            data["model_data"] = model_data
        if model_meta is not None:
            data["model_meta"] = model_meta
        if features is not None:
            data["features"] = features
        if table is not None:
            data["table"] = table
        if components is not None:
            for component in components:
                assert type(component) == Component, "`components`는 list(`sktmls.components.Component`) 형태여야 합니다."
            data["components"] = [component.name for component in components]

        updated_model_id = self._request(
            method="PUT", url=f"{MLS_MODELS_API_URL}/{manual_model.id}", data=data
        ).results.get("id")

        manual_model.reset(**self.get_manual_model(id=updated_model_id).get())
        return manual_model

    def delete_model(self, model: Union[AutoMLModel, ManualModel], delete_table: False) -> MLSResponse:
        """
        모델 삭제합니다.

        ## Args

        - model: (`sktmls.models.AutoMLModel` or `sktmls.models.ManualModel`) AutoML 혹은 일반 모델 객체
        - delete_table: (optional) (bool) manual_model의 경우 모델의 연결된 dynamodb table을 함께 지우는 옵션. (기본값: `False`)

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```
        ml_model_client.delete_model(model=model, delete_table=True)
        ```
        """

        assert type(model) in [
            AutoMLModel,
            ManualModel,
        ], "`model`은 `sktmls.models.AutoMLModel` 혹은 `sktmls.models.ManualModel` 타입이어야 합니다."

        response = self._request(method="DELETE", url=f"{MLS_MODELS_API_URL}/{model.id}")

        if delete_table and model.table:
            assert type(model) == ManualModel, "ManualModel의 경우에만 테이블 삭제 옵션을 사용할 수 있습니다"
            self._request(method="DELETE", url=f"{MLS_DYNAMODB_TABLES_API_URL}/{model.table}")

        return response

    def update_manual_model_meta(
        self,
        manual_model: ManualModel,
        model_meta: dict,
        partial_update: bool = False,
    ) -> ManualModel:
        """
        일반 모델의 모델 메타 정보를 수정합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체
        - model_meta: (dict) 수정할 모델 메타 정보
        - partial_update: (optional) (dict) 기존 모델 메타에 덮어씌울지(True) 기존 모델 메타를 교체할지(False) 여부 (기본값: `False`)

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```
        manual_model = ml_model_client.get_manual_model(name="my_manual_model", version="v1")
        manual_model = ml_model_client.update_manual_model_meta(
            manual_model=manual_model,
            model_meta={
                "model_lib": "xgboost",
                "model_lib_version": "1.2.0",
            },
            partial_update = True
        )
        manual_model_without_meta = ml_model_client.update_manual_model_meta(
            manual_model=manual_model,
            model_meta={}
        )

        ```
        """
        assert type(manual_model) == ManualModel

        if not partial_update:
            return self.update_manual_model(manual_model=manual_model, model_meta=model_meta)

        data = manual_model.model_meta
        data.update(model_meta)

        return self.update_manual_model(manual_model=manual_model, model_meta=data)

    def update_manual_model_components(
        self,
        manual_model: ManualModel,
        components: List[Component],
    ) -> ManualModel:
        """
        일반 모델의 컴포넌트 정보를 수정합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체
        - components: list(`sktmls.components.Component`) 수정할 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```
        component = component_client.get_component(name="test_component")
        manual_model = ml_model_client.get_manual_model(name="manual_test_model", version="v1")

        updated_manual_model = ml_model_client.update_manual_model_components(
            manual_model=manual_model,
            components=[component],
        )
        no_component_manual_model = ml_model_client.update_manual_model_components(
            manual_model=manual_model,
            components=[],
        )

        ```
        """
        assert type(manual_model) == ManualModel
        for component in components:
            assert type(component) == Component, "`components`는 list(`sktmls.components.Component`) 형태여야 합니다."

        return self.update_manual_model(manual_model=manual_model, components=components)

    def is_batch_prediction_ready(self, manual_model: ManualModel) -> bool:
        """
        일반 모델에 등록된 배치추론 테이블이 존재하고 업로드가 끝났는지 체크합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체

        ## Returns
        `bool`

        ## Example

        ```
        my_model = ml_model_client.get_manual_model(name="manual_test_model", version="v1")
        ml_model_client.manual_model_table_exists(manual_model=my_model)

        ```
        """
        assert type(manual_model) == ManualModel

        table = manual_model.table
        if not table:
            raise MLSClientError(code=404, msg="해당 일반모델에 배치추론 테이블이 연결되지 않았습니다.")

        dynamodb_tables = self._request(method="GET", url=MLS_DYNAMODB_TABLES_API_URL).results.get("tables")
        if table not in dynamodb_tables:
            return False

        status_cli = StatusClient(
            env=self._MLSClient__env,
            runtime_env=self._MLSClient__runtime_env,
            username=self._MLSClient__username,
            password=self._MLSClient__password,
        )
        running_jobs = [j["jobName"] for j in status_cli.get_batch_status("running").jobs]

        if f"{table}-loading" in running_jobs or f"{table}-deletion" in running_jobs:
            return False

        return True

    def is_online_prediction_ready(self, manual_model: ManualModel) -> bool:
        """
        일반 모델에 등록된 아티팩트(model.jobilb)이 존재하는지 확인합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체

        ## Returns
        `bool`

        ## Example

        ```
        my_model = ml_model_client.get_manual_model(name="manual_test_model", version="v1")
        ml_model_client.manual_model_artifact_exists(manual_model=my_model)

        ```
        """
        assert type(manual_model) == ManualModel

        data = {"model_name": manual_model.name, "model_version": manual_model.version}
        artifact = self._request(method="GET", url=MLS_MODEL_REGISTRY_ARTIFACTS_API_URL, params=data).results.get(
            "info"
        )

        if not artifact:
            return False
        return True



from typing import List

from dateutil import parser

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSResponse, MLSClientError

MLS_COMPONENTS_API_URL = "/api/v1/components"


class Component:
    """
    MLS 컴포넌트 클래스입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) 컴포넌트 고유 ID
            - name: (str) 컴포넌트 이름
            - info: (str) 컴포넌트 정보
            - is_latest: (bool) 최신여부
            - user: (str) 컴포넌트 생성 계정명
            - created_at: (datetime) 생성일시
            - updated_at: (datetime) 수정일시

        ## Returns
        `sktmls.components.Component`
        """
        self.id = kwargs.get("id")
        self.name = kwargs.get("name")
        self.info = kwargs.get("info")
        self.is_latest = kwargs.get("is_latest")
        self.user = kwargs.get("user")
        if kwargs.get("created_at"):
            self.created_at = parser.parse(kwargs.get("created_at"))
        if kwargs.get("updated_at"):
            self.updated_at = parser.parse(kwargs.get("updated_at"))

    def get(self):
        return self.__dict__

    def reset(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class ComponentClient(MLSClient):
    """
    MLS 컴포넌트 관련 기능들을 제공하는 클라이언트입니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
        username: str = None,
        password: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.components.ComponentClient`

        ## Example

        ```
        component_client = ComponentClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """

        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def create_component(self, name: str, info: str, is_latest: bool) -> Component:
        """
        컴포넌트를 생성합니다.

        ## Args

        - name: (str) 컴포넌트 이름
        - info: (str) 컴포넌트 정보
        - is_latest: (bool) 최신여부

        ## Returns
        `sktmls.components.Component`

        - id: (int) 컴포넌트 고유 ID
        - name: (str) 컴포넌트 이름
        - info: (str) 컴포넌트 정보
        - is_latest: (bool) 최신여부
        - user: (str) 컴포넌트 생성 계정명
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        component = component_client.create_component(
            name="my_component",
            info="my_info",
            is_latest=True,
        )
        ```
        """

        data = {"name": name, "info": info, "is_latest": is_latest}

        return Component(**self._request(method="POST", url=MLS_COMPONENTS_API_URL, data=data).results)

    def list_components(self, **kwargs) -> List[Component]:
        """
        컴포넌트 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 컴포넌트 조건
            - id: (int) 컴포넌트 고유 ID
            - name: (str) 컴포넌트 이름
            - is_latest: (bool) 최신여부

        ## Returns
        list(`sktmls.components.Component`)

        - id: (int) 컴포넌트 고유 ID
        - name: (str) 컴포넌트 이름
        - info: (str) 컴포넌트 정보
        - is_latest: (bool) 최신여부
        - user: (str) 컴포넌트 생성 계정명
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        components = component_client.list_components()
        ```
        """

        response = self._request(method="GET", url=f"{MLS_COMPONENTS_API_URL}", params=kwargs).results

        return [Component(**component) for component in response]

    def get_component(self, id: int = None, name: str = None, info: str = None, is_latest: bool = True) -> Component:
        """
        컴포넌트 정보를 가져옵니다. 여러개가 존재하는 경우 `is_latest`가 True인 최신의 컴포넌트를 반환합니다.

        ## Args: `id` 또는 `name` 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (optional) (int) 컴포넌트 고유 ID
        - name: (optional) (str) 컴포넌트 이름
        - is_latest: (optional) (bool) 최신여부 (기본값 : `True`)

        ## Returns
        `sktmls.components.Component`

        - id: (int) 컴포넌트 고유 ID
        - name: (str) 컴포넌트 이름
        - info: (str) 컴포넌트 정보
        - is_latest: (bool) 최신여부
        - user: (str) 컴포넌트 생성 계정명
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        component_by_id = component_client.get_component(
            id=3
        )
        component_by_name = component_client.get_component(
            name="my_component",
            is_latest=True
        )
        ```
        """
        assert id or name or info, "`id` 또는 `name` 또는 `info` 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        components = self.list_components(id=id, name=name, info=info, is_latest=is_latest)
        if len(components) == 0:
            raise MLSClientError(code=404, msg="컴포넌트가 없습니다.")
        elif len(components) > 1:
            raise MLSClientError(code=409, msg="컴포넌트가 여러개 존재합니다.")

        return Component(**self._request(method="GET", url=f"{MLS_COMPONENTS_API_URL}/{components[0].id}").results)

    def update_component(self, component: Component, is_latest: bool) -> Component:
        """
        컴포넌트 정보 중 `is_latest`를 수정합니다.

        ## Args

        - component: (`sktmls.components.Component`) 컴포넌트 객체
        - is_latest: (bool) 최신여부

        ## Returns
        `sktmls.components.Component`

        - id: (int) 컴포넌트 고유 ID
        - name: (str) 컴포넌트 이름
        - info: (str) 컴포넌트 정보
        - is_latest: (bool) 최신여부
        - user: (str) 컴포넌트 생성 계정명
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        component = component_client.get_component(name="hello")
        component = component_client.update_component(
            component=component,
            is_latest=True,
        )
        ```
        """
        assert type(component) == Component

        data = {
            "name": component.get().get("name"),
            "info": component.get().get("info"),
            "is_latest": is_latest,
        }

        component.reset(
            **self._request(method="PUT", url=f"{MLS_COMPONENTS_API_URL}/{component.id}", data=data).results
        )
        return component

    def delete_component(self, component: Component) -> MLSResponse:
        """
        컴포넌트 삭제합니다.

        ## Args

        - component: (`sktmls.components.Component`) 컴포넌트 객체

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```
        component_client.delete_component(component)
        ```
        """

        assert type(component) == Component

        return self._request(method="DELETE", url=f"{MLS_COMPONENTS_API_URL}/{component.id}")



from .mls_model import MLSModel, MLSLightGBMModel, MLSXGBoostModel, MLSRuleModel, MLSGenericModel, MLSModelError
from .ml_model import MLModelClient, MLModel, AutoMLModel, ManualModel, MLModelStatus

__all__ = [
    "MLSModel",
    "MLSLightGBMModel",
    "MLSXGBoostModel",
    "MLSRuleModel",
    "MLSGenericModel",
    "MLSModelError",
    "MLModelClient",
    "MLModel",
    "AutoMLModel",
    "ManualModel",
    "MLModelStatus",
]



from typing import List, Dict

from dateutil import parser

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSResponse, MLSClientError

MLS_FILTERS_API_URL = "/api/v1/filters"


class Filter:
    """
    MLS 필터 클래스입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) 필터 고유 ID
            - name: (str) 필터 이름
            - conditions: (list(dict)) 필터 조건 리스트
            - created_at: (datetime) 생성일시
            - updated_at: (datetime) 수정일시

        ## Returns
        `sktmls.filters.Filter`
        """
        self.id = kwargs.get("id")
        self.name = kwargs.get("name")

        if kwargs.get("body"):
            self.conditions = kwargs.get("body", {}).get("conditions", None)
        elif kwargs.get("conditions"):
            self.conditions = kwargs.get("conditions", None)

        if kwargs.get("created_at"):
            self.created_at = parser.parse(kwargs.get("created_at"))
        if kwargs.get("updated_at"):
            self.updated_at = parser.parse(kwargs.get("updated_at"))

    def get(self):
        return self.__dict__

    def reset(self, **kwargs):
        for k, v in kwargs.items():
            if k == "body":
                setattr(self, "conditions", v.get("conditions", None))
            else:
                setattr(self, k, v)


class FilterClient(MLSClient):
    """
    MLS 필터 관련 기능들을 제공하는 클라이언트입니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
        username: str = None,
        password: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.filters.FilterClient`

        ## Example

        ```
        filter_client = FilterClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """

        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def create_filter(self, name: str, conditions: List[Dict]) -> Filter:
        """
        필터를 생성합니다.

        ## Args

        - name: (str) 필터 이름
        - conditions: (list(dict)) 필터 조건 리스트
            - key: 프로파일 키
            - operator: 연산자 (`==` | `!=` | `<=` | `>=` | `<` |  `>`)
            - value: 비교할 값

        ## Returns
        `sktmls.filters.Filter`

        - id: (int) 필터 고유 ID
        - name: (str) 필터 이름
        - conditions: (list(dict)) 필터 조건 리스트
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        filter = filter_client.create_filter(
            name="hello",
            conditions=[
                {"key": "age", "operator": "<", "value": "19"},,
                {"key": "gender", "operator": "==", "value": "F"},
            ]
        )
        ```
        """

        data = {
            "name": name,
            "body": {"conditions": conditions},
        }

        return Filter(**self._request(method="POST", url=MLS_FILTERS_API_URL, data=data).results)

    def list_filters(self, **kwargs) -> List[Filter]:
        """
        필터 리스트를 가져옵니다. (필터 조건 제외)

        ## Args

        - kwargs: (optional) (dict) 필터 조건
            - id: (int) 필터 고유 ID
            - name: (str) 필터 이름

        ## Returns
        list(`sktmls.filters.Filter`)

        - id: (int) 필터 고유 ID
        - name: (str) 필터 이름
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        filters = filter_client.list_filters()
        ```
        """

        response = self._request(method="GET", url=f"{MLS_FILTERS_API_URL}", params=kwargs).results

        return [Filter(**filter) for filter in response]

    def get_filter(self, type: str = None, id: int = None, name: str = None) -> Filter:
        """
        필터 정보를 가져옵니다.

        ## Args: `id` 또는 `name` 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (optional) (int) 필터 고유 ID
        - name: (optional) (str) 필터 이름

        ## Returns
        `sktmls.filters.Filter`

        - id: (int) 필터 고유 ID
        - name: (str) 필터 이름
        - conditions: (list(dict)) 필터 조건 리스트
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        filter_by_id = filter_client.get_filter(
            id=3
        )
        filter_by_name = filter_client.get_filter(
            name="hello"
        )
        ```
        """
        assert id or name, "`id` 또는 `name` 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        filters = self.list_filters(type=type, id=id, name=name)
        if len(filters) == 0:
            raise MLSClientError(code=404, msg="필터가 없습니다.")
        elif len(filters) > 1:
            raise MLSClientError(code=409, msg="필터가 여러개 존재합니다.")
        return Filter(**self._request(method="GET", url=f"{MLS_FILTERS_API_URL}/{filters[0].id}").results)

    def update_filter(self, filter: Filter, name: str = None, conditions: List[Dict] = None) -> Filter:
        """
        필터 정보를 수정합니다.

        ## Args

        - filter: (`sktmls.filters.Filter`) 필터 객체
        - name: (str) 필터 이름
        - conditions: (list(dict)) 필터 조건 리스트
            - key: 프로파일 키
            - operator: 연산자 (`==` | `!=` | `<=` | `>=` | `<` |  `>`)
            - value: 비교할 값

        ## Returns
        `sktmls.filters.Filter`

        - id: (int) 필터 고유 ID
        - name: (str) 필터 이름
        - conditions: (list(dict)) 필터 조건 리스트
        - created_at: (datetime) 생성일시
        - updated_at: (datetime) 수정일시

        ## Example

        ```
        filter = filter_clinet.get_filter(name="hello")
        filter = filter_client.update_filter(
            filter=filter,
            name="bye",
            conditions=[
                {"key": "age", "operator": "<", "value": "50"},,
                {"key": "gender", "operator": "==", "value": "M"},
            ]
        )
        ```
        """
        assert type(filter) == Filter

        data = {"name": filter.get().get("name"), "body": {"conditions": filter.get().get("conditions")}}
        if name:
            data["name"] = name
        if conditions:
            data["body"] = {"conditions": conditions}

        filter.reset(**self._request(method="PUT", url=f"{MLS_FILTERS_API_URL}/{filter.id}", data=data).results)
        return filter

    def delete_filter(self, filter: Filter) -> MLSResponse:
        """
        필터 삭제합니다.

        ## Args

        - filter: (`sktmls.filters.Filter`) 필터 객체

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```
        filter_client.delete_filter(filter)
        ```
        """

        assert type(filter) == Filter

        return self._request(method="DELETE", url=f"{MLS_FILTERS_API_URL}/{filter.id}")



from typing import Any, Dict, List

from sktmls.models import MLSLightGBMModel, MLSXGBoostModel


class DefaultLightGBMModel(MLSLightGBMModel):
    """
    MLS 모델 레지스트리에 등록되는 LightGBM 기본 클래스입니다.

    온라인 추론에 없이 MLS 모델 레지스트리에 바이너리만을 등록하고자 할 때 사용합니다.
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        """
        본 클래스의 predict 함수는 아무런 일을 하지 않습니다.
        """
        return {"items": []}


class DefaultXGBoostModel(MLSXGBoostModel):
    """
    MLS 모델 레지스트리에 등록되는 XGBoost 기본 클래스입니다.

    온라인 추론에 없이 MLS 모델 레지스트리에 바이너리만을 등록하고자 할 때 사용합니다.
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        """
        본 클래스의 predict 함수는 아무런 일을 하지 않습니다.
        """
        return {"items": []}



from dateutil import parser
from enum import Enum

from sktmls import MLSClient, MLSENV, MLSRuntimeENV


class MLFeature:
    class Type(Enum):
        """
        AutoML 피쳐 타입입니다.

        - SCORE: 스코어 기반 고객 타겟팅
        - CLASSIFICATION: 분류
        - REGRESSION: 회귀
        """

        CATEGORY = "category"
        NUMBER = "number"

    class Status(Enum):
        """
        AutoML 피쳐 상태입니다.

        - CREATED: 생성됨
        - IN_SERVICE: 서비스 중
        - DEPRECATED: 더 이상 지원하지 않음
        """

        CREATED = "created"
        IN_SERVICE = "in_service"
        DEPRECATED = "deprecated"

    def __init__(self, **kwargs):
        """
        AutoML 피쳐 클래스.

        ## Attributes

        - id: (int) 피쳐 ID
        - name: (str) 이름
        - source: (str) 원천 경로
        - type: (`sktmls.ml_features.MLFeature.Type`) 타입
        - status: (`sktmls.ml_features.MLFeature.Status`) 서비스 상태
        - title: (str) 타이틀
        - description: (str) 설명
        - created_at: (datetime.datetime) 생성 시각
        - updated_at: (datetime.datetime) 갱신 시각
        """
        self.id = kwargs.get("id")
        self.name = kwargs.get("name")
        self.source = kwargs.get("source")
        self.type = MLFeature.Type(kwargs.get("type"))
        self.type = MLFeature.Status(kwargs.get("status"))
        self.title = kwargs.get("title")
        self.description = kwargs.get("description")

        try:
            self.created_at = parser.parse(kwargs.get("created_at"))
        except TypeError:
            self.created_at = None

        try:
            self.updated_at = parser.parse(kwargs.get("updated_at"))
        except TypeError:
            self.updated_at = None

    def __str__(self) -> str:
        return f"{self.source}.{self.name}"


class MLFeatureClient(MLSClient):
    def __init__(
        self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, username: str = None, password: str = None
    ):
        """
        AutoML 피쳐 관련 기능을 제공하는 클라이언트 클래스입니다.

        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.datasets.DatasetClient`

        ## Example

        ```
        client = DatasetClient(env=MLSENV.STG, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def create_feature(self, name: str, source: str, type: MLFeature.Type, title: str = None, description: str = None):
        """
        새로운 AutoML 피쳐를 생성합니다.

        ## Args

        - name: (str) 이름
        - source: (str) 원천 경로
        - type: (`sktmls.ml_features.MLFeature.Type`) 타입
        - title: (optional) (str) 타이틀
        - description: (optional) (str) 설명

        ## Returns
        `sktmls.ml_features.MLFeature`
        ```
        """
        data = {
            "name": name,
            "source": source,
            "type": type.value,
        }
        if title is not None:
            data["title"] = title
        if description is not None:
            data["description"] = description
        return self._request(method="POST", url="api/v1/ml_features", data=data)



from typing import Any, Dict, List

from sktmls.models import MLSModelError, MLSRuleModel


class MbrVipInfoRuleModel(MLSRuleModel):
    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if len(self.features) != len(x):
            raise MLSModelError("The length of input is different from that of features in model.json")

        mbr_card_gr_cd = x[0]
        mbr_use_discount_amt_cum = x[1]

        if mbr_card_gr_cd != "V" and mbr_use_discount_amt_cum > 0:
            return {"items": [{"id": "INF0000009", "name": "VIP 혜택 및 승급 방법 안내", "type": "mbr_vip_info", "props": {}}]}
        else:
            return {"items": []}
